/*
 * hrtime.c
 *
 * Copyright (C) 2000, Nick Rasmussen <nick@jive.org>
 *
 */
#include "hrtime.h"

#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>


static struct hrtime_struct *current = 0;
static int num_structs_mapped = 0;

#define HRTIME_PATH_LENGTH 256

int
get_hrtime_struct(pid_t pid, struct hrtime_struct **dest)
{
    char path[HRTIME_PATH_LENGTH];
    int  fd, count;

    if (dest == 0)
	return -1;

    if (current == 0)
    {
	fd = open("/proc/self/hrtime", O_RDONLY);
	if (fd < 0)
	    return fd;

	current = (struct hrtime_struct *) mmap(0, sizeof(struct hrtime_struct),
						PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	if (current == MAP_FAILED)
	{
	    current = 0;
	    return errno;
	}
    }

    if (pid != 0 && pid != getpid())
    {
	count = snprintf(path, HRTIME_PATH_LENGTH, "/proc/%d/hrtime", pid);
	if (count < 0 || count >= HRTIME_PATH_LENGTH)
	    return -1;
	
	fd = open(path, O_RDONLY);
	if (fd < 0)
	    return fd;
	
	*dest = (struct hrtime_struct *) mmap(0, sizeof(struct hrtime_struct),
					      PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	
	if (current == MAP_FAILED)
	{
	    *dest = 0;
	    return errno;
	}
    }
    else
    {
	*dest = current;
    }

    num_structs_mapped += 1;
    return 0;
} 

int
free_hrtime_struct(struct hrtime_struct *hr)
{
    int error;

    if (hr == 0) return 0;
    
    if (hr != current)
    {
	error = munmap((void *) hr, sizeof(struct hrtime_struct));
	if (error != 0)
	    return error;
    }

    num_structs_mapped -= 1;

    if (num_structs_mapped == 0)
    {
	error = munmap((void *) current, sizeof(struct hrtime_struct));
	if (error != 0)
	    return error;
	current = 0;
    }

    return 0;
}

#define HRTIME_CYCLE_LIMIT 100

void
get_hrtime(struct hrtime_struct *hr, hrtime_t *dest)
{
#if HRTIME_HAS_OFFSETS
    htrime_t tmp;
    long     offset;

    do {
	get_current_hrtime(dest);
	offset = current->offset_to_cpu0;
	get_current_hrtime(&tmp);
    }
    while (tmp - dest > HRTIME_CYCLE_LIMIT);

    /* start_time + offset is start_time relative to the cpu we are currently
     * executing on
     */
    *dest -= (hr->start_time + offset);
#else
    get_current_hrtime(dest);
    *dest -= hr->start_time;
#endif
}

void
get_hrvtime(struct hrtime_struct *hr, hrtime_t *dest)
{
    hrtime_t ld, vtime, tmp;

#if HRTIME_HAS_OFFSETS
    long     my_off, hr_off;
#endif

    do {
	get_current_hrtime(dest);

	ld        = hr->last_dispatch;
	vtime     = hr->vtime;
#if HRTIME_HAS_OFFSETS
	my_off    = current->offset_to_cpu0;
	hr_off    = hr->offset_to_cpu0;
#endif
	get_current_hrtime(&tmp);
    }
    while (ld == hr->last_dispatch && tmp - *dest > HRTIME_CYCLE_LIMIT);

    if (ld != 0)
    {
#if HRTIME_HAS_OFFSETS
	/* make ld in reference to our cpu */
	ld += (hr_off - my_off);
#endif
	*dest -= ld;
	*dest += vtime;
    }
    else
    {
	*dest = vtime;
    }
}

void
get_hrutime(struct hrtime_struct *hr, hrtime_t *dest)
{
    hrtime_t ld, utime, tmp;
    long     in_system;

#if HRTIME_HAS_OFFSETS
    long     my_off, hr_off;
#endif

    do {
	get_current_hrtime(dest);

	ld        = hr->last_us_dispatch;
	utime     = hr->utime;
	in_system = hr->in_system;
#if HRTIME_HAS_OFFSETS
	my_off    = current->offset_to_cpu0;
	hr_off    = hr->offset_to_cpu0;
#endif
	get_current_hrtime(&tmp);
    }
    while (ld == hr->last_us_dispatch && tmp - *dest > HRTIME_CYCLE_LIMIT);

    if (ld != 0 && in_system == 0)
    {
#if HRTIME_HAS_OFFSETS
	/* make ld in reference to our cpu */
	ld += (hr_off - my_off);
#endif
	*dest -= ld;
	*dest += utime;
    }
    else
    {
	*dest = utime;
    }

}

void
get_hrstime(struct hrtime_struct *hr, hrtime_t *dest)
{
    hrtime_t ld, stime, tmp;
    long     in_system;

#if HRTIME_HAS_OFFSETS
    long     my_off, hr_off;
#endif

    do {
	get_current_hrtime(dest);

	ld        = hr->last_us_dispatch;
	stime     = hr->stime;
	in_system = hr->in_system;
#if HRTIME_HAS_OFFSETS
	my_off    = current->offset_to_cpu0;
	hr_off    = hr->offset_to_cpu0;
#endif
	get_current_hrtime(&tmp);
    }
    while (ld == hr->last_us_dispatch && tmp - *dest > HRTIME_CYCLE_LIMIT);

    if (ld != 0 && in_system != 0)
    {
#if HRTIME_HAS_OFFSETS
	/* make ld in reference to our cpu */
	ld += (hr_off - my_off);
#endif
	*dest -= ld;
	*dest += stime;
    }
    else
    {
	*dest = stime;
    }
}
